# Copyright (c) 2018 Trevor Taylor
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import os
import mimetypes

from xju.xn import first_line_of as l1,in_context
from xju import pq
from xju.misc import fromJson,toJson
from .wsgi import getVariablesFromWSGIenviron, getCookiesFromWSGIenviron, getHTTPHeadersFromWSGIenviron
from typing import Set,Callable,Dict,Union,Tuple,List
from . import public_functions, restricted_functions, ClientError,Forbidden,NotFound,Response,promoteContent

def getParam(param_name,
             json_params,
             request_params,
             request_attrs,
             param_defaults):
    try:
        try:
            if param_name=='json_params':
                return json_params
            if param_name in json_params:
                return json_params.get(param_name)
            if param_name in request_params:
                return request_params.get(param_name)
            if param_name in request_attrs:
                return request_attrs.get(param_name)
            return param_defaults[param_name]
        except KeyError as e:
            raise Exception(f'unknown parameter {param_name}')
        pass
    except:
        json_param_names=json_params.keys()
        request_param_names=request_params.keys()
        request_attr_names=request_attrs.keys()
        param_default_names=param_defaults.keys()
        raise in_context('get value of %(param_name)s from params supplied as json-encoded "json_params" HTTP param (%(json_param_names)s), HTTP params (%(request_param_names)s), webapp2 request attributes (%(request_attr_names)s) or function paramter defaults (%(param_default_names)s)'%vars()) from None
    pass

def makeParams(remote_addr,method,headers,params,url,referer,cookies,f):
    'make dictionary of params for calling function %(f)s, getting them from remote_addr, method, headers, GET/POST params, url, cookies'''
    try:
        json_params=fromJson(params.get('json_params','{}'))
        param_names=f.__code__.co_varnames[0:f.__code__.co_argcount]
        func_defaults=f.__defaults__ or []
        no_default_params=len(param_names)-len(func_defaults)
        param_defaults=dict([ (param_name,default_value) for param_name,default_value in zip(param_names[no_default_params:],func_defaults)])
        #request.__dict__ does not give us anything useful
        request_attrs={
            'params':params,
            'cookies':cookies,
            'headers':headers,
            'method':method,
            'referer':referer,
            'url':url,
            'remote_addr':remote_addr,
            }
        result=dict([ (_,getParam(_,
                                  json_params,
                                  params,
                                  request_attrs,
                                  param_defaults)) for _ in param_names ])
        return result
    except:
        raise in_context(l1(makeParams.__doc__)%vars()) from None
    pass

class Dispatcher:
    def __init__(self,app,dir,
                 log=lambda s: print('INFO: {s}'.format(**vars())),
                 logError=lambda s: print('ERROR: {s}'.format(**vars()))):
        '''serve pages from {app}'s functions and {dir}'s files'''
        '''(see self.main)'''
        self.app=app
        self.dir=dir
        self.log=log
        self.logError=logError
        pass
    def main(self,environ,start_response):
        '''dispatch request descibed by WSGI environ
             - satisfies request by corresponding names function in self.app
             - fallback is static file from self.dir
           (WSGI application function)'''
        try:
            if len(environ['PATH_INFO'])>1:
                assert environ['PATH_INFO'].startswith('/') #wsgi always does
                path=environ['PATH_INFO'][1:]
            else:
                path='index.html'
                pass
            return \
                self.dispatchToFunction(path,environ,start_response,self.app) or \
                self.dispatchToFile(path,start_response) or \
                self.notFound(path,start_response)
        except Forbidden as e:
            self.log('INFO: '+str(
                in_context(l1(Dispatcher.main.__doc__).format(**vars()))))
            return self.forbidden(start_response)
        except NotFound as e:
            self.log(str(
                in_context(l1(Dispatcher.main.__doc__).format(**vars()))))
            return self.notFound(path,start_response)
        except ClientError as e:
            return self.clientError(start_response,e)
        except Exception as e:
            self.logError('ERROR: '+str(
                in_context(l1(Dispatcher.main.__doc__).format(**vars()))))
            return self.serverError(start_response)
        pass
    def dispatchToFunction(self,name,environ,start_response,mod):
        'dispatch {name} to function in {mod} a name->function dictionary'
        'e.g. app can be a module'
        try:
            if '/' in name:
                m,n=name.split('/',1)
                if m in mod.__dict__:
                    return self.dispatchToFunction(n,environ,start_response,mod.__dict__[m])
                else:
                    return None
                pass
            fname=name.replace('-','_').replace(' ','_').replace('.','_')
            f=mod.__dict__.get(fname,None)
            if not f: return None
            if not callable(f):
                raise Forbidden('%(fname)s is not callable ie not a function like "def %(fname)s():"'%vars())
            if not f in public_functions and not f in restricted_functions:
                raise Forbidden('%(fname)s is not annotated with @wal.public (or @public) or @wal.restricted (or @restricted)'%vars())
            result=None
            url=(environ.get('wsgi.url_scheme')+'://'+
                 environ.get('HTTP_HOST',
                             environ.get('SERVER_NAME')+':'+
                             environ.get('SERVER_PORT'))+
                 environ.get('SCRIPT_NAME','')+
                 environ.get('PATH_INFO',''))
            cookies=getCookiesFromWSGIenviron(environ)
            if f in restricted_functions:
                result=restricted_functions[f](url,cookies)
                pass
            if not result:
                params=getVariablesFromWSGIenviron(environ)
                headers=getHTTPHeadersFromWSGIenviron(environ)
                method=environ['REQUEST_METHOD'] #e.g. 'GET', 'POST'
                referer=environ.get('HTTP_REFERER',None)
                remote_addr=environ['REMOTE_ADDR']
                result=f(**makeParams(remote_addr,
                                      method,
                                      headers,
                                      params,
                                      url,
                                      referer,
                                      cookies,
                                      f))
                pass
            self.log('INFO: {name} used {mod.__name__}.{fname}()'.format(**vars()))
            result=promoteContent(result)
            headers=result.cookieHeaders()
            if result.location:
                headers.append( ('Location',result.location) )
                start_response('303 See Other',headers)
                return [''.encode('utf-8')]
            headers.extend([(n,v) for n,v in result.headers
                            if not n in ('Content-Type','Content-Encoding')])
            c=dict([ (n,v) for n,v in result.headers
                     if n in ('Content-Type','Content-Encoding')])
            if result.contentType: c['Content-Type']=result.contentType
            if result.contentEncoding: c['Content-Encoding']=result.contentEncoding
            headers.extend(c.items())
            start_response('200 OK',headers)
            return [result.content]
        except:
            raise in_context(l1(Dispatcher.dispatchToFunction.__doc__).format(
                **vars())) from None
        pass
    
    def dispatchToFile(self,path,start_response):
        'dispatch {self.dir}/{path} if it exists as a plain file'
        x=os.path.join(self.dir,path)
        if os.path.exists(x):
            t=mimetypes.guess_type(path,False)
            headers=[('Content-Type',t[0] or 'text/plain')]
            if t[1]:
                headers.append(('Content-Encoding',t[1]))
                pass
            with open(x,'rb') as f:
                start_response('200 OK', headers)
                return [f.read()]
            pass
        return None
    
    def notFound(self,path,start_response):
        self.log('INFO: {path} not found'.format(**vars()))
        start_response('404 Not Found', 
                       [('Content-Type', 'text/plain; charset=UTF-8')])
        return ['Not Found'.encode('utf-8')]
    
    def forbidden(self,start_response):
        start_response('403 Forbidden', 
                       [('Content-Type', 'text/plain; charset=UTF-8')])
        return ['Forbidden'.encode('utf-8')]
    
    def serverError(self,start_response):
        start_response('500 Internal Server Error', 
                       [('Content-Type', 'text/plain; charset=UTF-8')])
        return ['Server Error'.encode('utf-8')]
    
    def clientError(self,start_response,e):
        start_response('400 Bad Request', 
                       [('Content-Type', 'text/plain; charset=UTF-8')])
        return [Exception.__str__(e).encode('utf-8')]
    
    pass





# coding: utf-8

# Copyright (c) 2018 Trevor Taylor
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import urllib
from xju.xn import in_context,in_function_context
from xju.xn import first_line_of as l1
from xju.assert_ import Assert
from typing import Tuple,Dict,Mapping,Union,List,Any
from urllib.parse import unquote as urlunquote

def parseHeaders(mimePart:bytes)->Tuple[Dict[str,str],bytes]:
    try:
        Assert(mimePart).isInstanceOf(bytes)
        h,rest=mimePart.split(b'\r\n\r\n', 1)
        headers=h.split(b'\r\n',1)
        headers2=[_.split(b':',1) for _ in headers]
        headers3=[(_[0].decode('utf-8'), _[1].strip().decode('utf-8'))
                 for _ in headers2] #REVISIT: not utf-8 but what?
        return dict(headers3), rest
    except:
        mps=mimePart[0:256]
        raise in_context('parse headers from mime part %(mps)s...'%vars()) from None
    pass

def parseQuoted(x:str):
    assert x.startswith('"'), x
    assert x.endswith('"'), x
    return x[1:-1].replace(r'\"','"')

def parseDisposition(dispositionValue:str):
    '''Parse Content-Disposition value {dispositionValue}'''
    try:
        assert dispositionValue.strip().startswith('form-data;'), dispositionValue
        v=dispositionValue.split(';')[1:]
        v2=[_.strip().split('=',1) for _ in v]
        v3=[(_[0],parseQuoted(_[1])) for _ in v2]
        return dict(v3)
    except:
        raise in_function_context(parseDisposition,vars()) from None
    pass

class FileVar(object):
    def __init__(self, filename:str,contentType:str, content:bytes):
        Assert(content).isInstanceOf(bytes)
        self.filename:str=filename
        self.contentType:str=contentType
        self.content:bytes=content
        pass
    pass

StrOrFileVar=Union[str,
                   FileVar, # for each part of multi-part/form-data
                   Tuple[int, Any]]     #  (length, wsgi.input) for application/octet-stream

def getVariablesFromWSGIenviron(wsgiEnv:Mapping[str,Any])->Dict[str,Union[str,FileVar,Tuple[int,Any],List[StrOrFileVar]]]:
    '''parse query string and wsgi.input into a dictionary from WSGI environ {wsgiEnv}'''
    '''like { varName : str or FileVar or list of str or FileVar }, e.g.:'''
    '''  { "id":"883999", "colours":["red","blue"] }'''
    '''... note list is only used where var has multiple values'''
    '''- where Content-Type is application/octet-stream, result dictiony'''
    '''  includes "body", a tuple ( content_length, file )'''
    '''  where content length is from Content-Length header (or -1) and'''
    '''  file is a file-like readable that presents the body of the request'''
    try:
        result:Dict[str,Union[str,FileVar,Tuple[int,Any],List[StrOrFileVar]]]={}
        d:List[Tuple[str,StrOrFileVar]]=[]
        if wsgiEnv.get('QUERY_STRING',''):
            nvs=[_.split('=',1) for _ in wsgiEnv['QUERY_STRING'].split('&')]
            d.extend(
                [ (_[0], urlunquote(_[1].replace('+',' ')))
                  for _ in nvs])
            pass
        if wsgiEnv['REQUEST_METHOD']!='GET' and 'CONTENT_TYPE' in wsgiEnv:
            #application/javascript; charset=utf-8
            ct:str=wsgiEnv['CONTENT_TYPE']
            encoding='iso-8859-1'
            if 'charset=' in ct:
                encoding=ct.split('charset=')[1].split(';')[0]
                pass
            if 'application/x-www-form-urlencoded' in ct:
                l=int(wsgiEnv.get('CONTENT_LENGTH',-1))
                y=wsgiEnv['wsgi.input'].read(l).decode(encoding)
                nvs=[_.split('=') for _ in y.split('&')]
                d.extend(
                    [ (_[0], urllib.parse.unquote(_[1].replace('+',' ')))
                      for _ in nvs])
            elif 'multipart/form-data' in ct:
                assert 'boundary=' in ct,ct
                # standard adds extra '--' in front of all occurrances
                # of bondary (and after last too)
                boundary:bytes=('\r\n--'+
                                ct.split('boundary=')[1]).encode('utf-8')
                body:bytes=b'\r\n'+wsgiEnv['wsgi.input'].read(
                    int(wsgiEnv.get('CONTENT_LENGTH',-1)))
                all:List[bytes]=body.split(boundary)[1:-1]
                for i,x in enumerate(all):
                    try:
                        assert x.startswith(b'\r\n'), x[0:256]
                        x=x[2:]
                        headers,rest=parseHeaders(x)
                        disposition=parseDisposition(
                            headers['Content-Disposition'])
                        if 'filename' in disposition:
                            fileType=headers.get('Content-Type','')
                            d.append( (disposition['name'], 
                                       FileVar(disposition['filename'],
                                               fileType,
                                               rest)))
                        else:
                            d.append( (disposition['name'],rest.decode('utf-8')) )
                            pass
                    except:
                        raise in_context('parse part %(i)s' % vars()) from None
                    pass
                pass
            elif 'application/octet-stream' in ct:
                length=int(wsgiEnv.get('CONTENT_LENGTH',-1))
                print(f'application/octet-stream, length {length}')
                
                d.append( ('body', (length, wsgiEnv['wsgi.input'])) )
            else:
                if int(wsgiEnv.get('CONTENT_LENGTH',0)):
                    raise Exception(
                        'unimplemented CONTENT_TYPE: '+wsgiEnv['CONTENT_TYPE']+' for POST with non-zero CONTENT_LENGTH')
            pass
        for key, value in d:
            if key in result:
                v=result[key]
                if isinstance(v,list):
                    v.append(value)
                else:
                    result[key]=[v,value]
                    pass
                pass
            else:
                result[key]=value
                pass
            pass
        return result
    except:
        raise in_function_context(getVariablesFromWSGIenviron,vars()) from None
    pass

def getCookiesFromWSGIenviron(environ:Mapping[str,str])->Dict[str,str]:
    '''get cookies from WSGI environ {environ} HTTP_COOKIE item as dictionary'''
    '''like { cookieName : str ] }, e.g.:'''
    '''  { "emailAddress" : "fred@dot.com" } '''
    try:
        result={}
        e=environ.get('HTTP_COOKIE',None)
        if e: result=dict([_.strip().split('=') for _ in e.split(';')])
        return result
    except:
        raise in_function_context(getCookiesFromWSGIenviron,vars()) from None
    pass


def getHTTPHeadersFromWSGIenviron(environ)->Dict[str,str]: #name,value
    '''get HTTP_x headers from WSGI environ {environ!r} as dictionary'''
    '''like { headerName : str }, e.g.:'''
    try:
        return dict([(name,value) for name,value in environ.items()
                     if name.startswith('HTTP_')])
    except:
        raise in_function_context(getHTTPHeadersFromWSGIenviron,vars()) from None
    pass



# Copyright (c) 2018 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import os
import sys
from typing import Set,Callable,Dict,Union,Tuple,List
from xju import rfc2616
from xju.xn import in_context,in_function_context
from xju.misc import toJson

from xju import jsonschema
from xju import pq
from xju import xn
    
public_functions:Set[str]=set()

restricted_functions:Dict[str,Callable]={}

# decorator - see test/app.py for use
def public(f):
    'make f available via all Dispatchers to web clients (eg browsers)'
    public_functions.add(f)
    return f

# decorator - see test/app.py for use
def restricted(authorizer):
    'make f available via all Dispatchers to web clients (eg browsers) if authorizer(url,cookies) returns None'
    def result(f):
        restricted_functions[f]=authorizer
        return f
    return result

# Response convenience functions - see test/app.py for use
def plainText(text):
    result=Response(content=text.encode('utf-8'),
                    contentType='text/plain; charset=UTF-8')
    return result

def html(text):
    result=Response(content=text.encode('utf-8'),
                    contentType='text/html; charset=UTF-8')
    return result

def cookie(name,value,path='/'):
    result=Response()
    result.cookies={
        validateCookieName(name): (validateCookieValue(value),{
            'Path':validateCookiePath(path)})}
    return result

def header(name,value):
    '''HTTP header'''
    result=Response()
    result.headers=[(name,value)]
    return result

def redirect(location:str,**etc):
    '''HTTP redirect to %(location)s, with extras %(etc)r'''
    cookies=etc.get('cookies',{})
    result=Response()
    result.cookies=cookies
    result.location=location
    return result

class ClientError(Exception):
    'error to pass back to client'
    def __init__(self,s):
        Exception.__init__(self,s)
        self.error=s
        pass
    pass

class Forbidden(Exception):
    def __init__(self,description):
        Exception.__init__(self,description)
        pass
    pass

class NotFound(Exception):
    def __init__(self,description):
        Exception.__init__(self,description)
        pass
    pass

class Response:
    def __init__(self,
                 content:Union[bytes, #pre-encoded
                               None]=None, #REVISIT: rules
                 contentType:Union[str,None]=None, # e.g. text/html; charset=UTF-8
                 contentEncoding:Union[str,None]=None):
        assert content is None or isinstance(content,bytes),type(content)
        assert contentType is None or content is not None, (contentType,type(content))
        assert contentEncoding is None or content is not None, (contentEncoding,type(content))
        self.content:Union[bytes,None]=content
        self.contentType:Union[str,None]=contentType
        self.contentEncoding:Union[str,None]=contentEncoding
        self.cookies:Dict[str,Tuple[str,Dict[str,str]]]={} # name : (value, {attrName:attrValue} )
        self.headers:List[Tuple[str,str]]=[] # (name,value)
        self.location:Union[None,str]=None # for redirect
        pass
    def __str__(self):
        c=(self.content or '')[0:150]
        if len(self.content or '')>150: c=c+'...'
        h=self.headers
        k=self.cookies
        return 'content {c!r}, headers {h!r}, cookies {k!r}'.format(**vars())
    def __add__(self,rhs_):
        'add {rhs_} to Response {self}'
        try:
            rhs=promoteContent(rhs_)
            assert isinstance(rhs,Response),type(rhs)
            lhs=self
            assert lhs.content is None or rhs.content is None, 'content specified more than once'
            result=Response()
            result.content=rhs.content or lhs.content
            result.contentType=rhs.contentType or lhs.contentType
            result.contentEncoding=rhs.contentEncoding or lhs.contentEncoding
            result.cookies=lhs.cookies.copy()
            result.cookies.update(rhs.cookies)
            result.headers=lhs.headers+rhs.headers
            result.location=rhs.location or lhs.location
            return result
        except:
            raise in_function_context(Response.__add__,vars()) from None
        pass
    def __radd__(self,lhs_):
        'add Response {self} to {lhs_}'
        try:
            return promoteContent(lhs_).__add__(self)
        except:
            raise in_function_context(Response.__radd__,vars()) from None
        pass
    def cookieHeaders(self):
        '''return headers for {self.cookies}'''
        try:
            result=[] # ('Set-Cookie', xxxx }
            for name,va in self.cookies.items():
                value=va[0]+''.join(['; {an}={av}'.format(**vars())
                                     for an,av in va[1].items()])
                result.append( ('Set-Cookie',f'''{name}={value}''') )
                pass
            return result
        except:
            raise in_function_context(Response.cookieHeaders,vars()) from None
        pass
    pass

def validateCookieName(name:str):
    '''validate RFC6265 cookie name {name!r}'''
    try:
        return rfc2616.validateToken(name)
    except:
        raise in_function_context(validateCookieName,vars()) from None
    pass

def validateCookieValue(v):
    '''validate RFC6265 cookie-value {v!r}'''
    try:
        if v.startswith('"') and not v.endswith('"'):
            raise Exception('{v!r} startswith double quote but does not end with double quote'.format(**vars()))
        x=v[1:-1] if v.startswith('"') else v
        for i,c in enumerate(x):
            o=ord(c)
            if o<0x21 or \
               (o>0x21 and o<0x23) or \
               (o>0x2b and o<0x2d) or \
               (o>0x3a and o<0x3c) or \
               (o>0x4b and o<0x5d) or \
               (o>0x7e):
                rest=result[i:]
                raise Exception('invalid character at {rest!r}'.format(**vars()))
            pass
        return v
    except:
        raise in_function_context(validateCookieValue,vars()) from None
    pass

def validateCookiePath(v):
    '''validate RFC6265 cookie Path value {v!r}'''
    try:
        for i,c in enumerate(v):
            try:
                if c in rfc2616.CTLs:
                    raise Exception('{c!r} is a RFC2616 control character'.format(**vars()))
                if c==';':
                    raise Exception('{c!r} is a semicolon'.format(**vars()))
            except:
                rest=v[i:]
                raise in_context('validate at ...{rest!r}'.format(**vars())) from None
            pass
        return v
    except:
        raise in_function_context(validateCookiePath,vars()) from None
    pass


def promoteContent(content:Union[Response, #already good
                                 pq.Selection, #text/html
                                 dict,bool,list,float]): #text/json REVISIT: rules
    '''promote content object to a valid Response'''
    contentType=type(content)
    try:
        if isinstance(content,Response):
            return content
        if isinstance(content,pq.Selection):
            return Response(content.utf8(),
                            'text/html; charset=UTF-8')
        if isinstance(content,dict) and (
            'result' in content or
            'error' in content):
            return Response(toJson(content).encode('utf-8'),
                            'text/json; charset=UTF-8')
        if content is None or \
           isinstance(content,dict) or isinstance(content,list) or \
           isinstance(content,int) or isinstance(content,float):
            return Response(toJson({'result':content}).encode('utf-8'),
                            'text/json; charset=UTF-8')
        raise Exception('do not know what HTTP HTTP CONTENT-TYPE to use for a {contentType} object - return an explicit wal.Response to set CONTENT-TYPE'.format(**vars()))
    except:
        raise in_function_context(promoteContent,vars()) from None
    pass



# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# CMC - Context Manager Complete/Crap
#
import asyncio
import sys
import contextlib
from typing import TypeVar, Iterable, Dict as _Dict, overload, Tuple, Sequence, Union, Optional
from typing import ItemsView, KeysView
from typing import Mapping, Type, List, Generic, Any
from collections import OrderedDict
from collections.abc import Coroutine
import builtins

from xju.xn import in_function_context
from xju.assert_ import Assert

T = TypeVar('T', bound=contextlib.AbstractContextManager)

def cmclass(cls:Type[T]) -> Type[T]:
    '''Class decorator that adds context management __enter__ and __exit__
       that enter and exit all type-hinted attributes implementing contextlib.AbstractContextManager
       then any existing __enter__.
       Note that to satisfy mypy, the decorated class must already
       implement contextlib.AbstractContextManager and the preferred way to do that is
       to inherit from xju.cmc.CM - see cmc.py.test for examples.
    '''
    base_classes_to_enter = [ base_class for base_class in cls.__bases__
                              if issubclass(base_class, contextlib.AbstractContextManager) ]
    attrs_to_enter = [ n for n, t in cls.__annotations__.items()
                         if is_subclass(n, t, contextlib.AbstractContextManager)]
    # need a unique place to keep resources acquired by enter so exit can
    # exit them, only want resources for self, not subclasses (which do their own
    # handling). Note replacing . with _ could lead to clashes but there's no
    # way to avoid all clashes.
    resources_attr_name=f'{cls.__module__}.{cls.__name__}'.replace('.','_')
    orig_enter=cls.__dict__.get('__enter__',lambda self: self)
    def enter(self,
              base_classes_to_enter=base_classes_to_enter,
              attrs_to_enter=attrs_to_enter,
              resources_attr_name=resources_attr_name,
              orig_enter=orig_enter) -> Type[T]:
        with contextlib.ExitStack() as tentative:
            for base_class in base_classes_to_enter:
                if base_class is not CM:
                    cm = __ClassCm(base_class, self)
                    tentative.enter_context(cm)
                    pass
                pass
            for n in attrs_to_enter:
                tentative.enter_context(getattr(self, n))
                pass
            assert getattr(self,resources_attr_name, None) is None, f'{cls.__module__}.{cls.__name__} has a f{resources_attr_name} attribute already. Perhaps it is inherited multiple times (which is not allowed by xju.cmc.cmclass decorator)?'
            orig_enter(self)
            setattr(self,resources_attr_name,tentative.pop_all())
        return self
    orig_exit=cls.__dict__.get('__exit__',lambda self,t,e,b: None)
    def exit(self,t,e,b,resources_attr_name=resources_attr_name) -> None:
        resources = getattr(self,resources_attr_name)
        setattr(self,resources_attr_name,None)
        with contextlib.closing(resources):
            return orig_exit(self,t,e,b)
    old_sa=cls.__setattr__
    def sa(self, name, value, attrs_to_enter=attrs_to_enter):
        if getattr(self,resources_attr_name,None) and name in attrs_to_enter:
            assert False, f'xju.cmc {self} has been entered so cannot replace context-managed attribute {cls.__module__}.{cls.__name__}.{name}'
        return old_sa(self,name,value)
    setattr(cls, '__enter__', enter)
    setattr(cls, '__exit__', exit)
    setattr(cls, '__setattr__', sa)
    cls.__annotations__['__enter__'] = type(enter)
    cls.__annotations__['__exit__'] = type(exit)
    return cls

class CM(contextlib.AbstractContextManager):
    def __enter__(self):
        assert False, f'{CM.__module__}.CM.__enter__ not overridden, have you forgotten @cmclass on {self.__module__}.{self.__class__.__name__}?'
        return self
    def __exit__(self, t, e, b) -> None:
        assert False, f'{CM.__module__}.CM.__exit__ not overridden, have you forgotten @cmclass on {self.__module__}.{self.__class__.__name__}?'
        pass

K = TypeVar('K')
V = TypeVar('V', bound=contextlib.AbstractContextManager)

class Dict(Mapping[K, V], contextlib.AbstractContextManager):
    '''dictionary that enters and exits its values, all of which must be context managers
       note that values may be added and removed before and/or after self is entered:
         before: the value will be entered when self is entered
         after:  the value will be entered before it is added (noting that any existing
                 value will be exited once (if) it is removed) - the enter-new-exit-old 
                 order might be important to you i.e. might not be what you need!'''
    entered = False
    x:_Dict[K,V]

    @overload
    def __init__(self):
        '''new empty Dict'''

    @overload
    def __init__(self, x:_Dict[K,V]):
        '''initialise with value from {x} assuming those values have not been "entered"'''

    @overload
    def __init__(self, x:Iterable[Tuple[K,V]]):
        '''initialise with value from {x} assuming those values have not been "entered"'''

    def __init__(self, *args, **kwargs):
        self.x = OrderedDict(*args, **kwargs)
        pass
    
    def __enter__(self):
        '''"enters" all values in order they were inserted'''
        with contextlib.ExitStack() as tentative:
            for k,v in self.x.items():
                tentative.enter_context(v)
            tentative.pop_all()
            pass
        self.entered = True
        return self

    def __exit__(self, t, e, b):
        '''"exits" all values in order they were inserted'''
        if self.entered:
            with contextlib.ExitStack() as resources:
                for k,v in self.x.items():
                    resources.push(v)
                    pass
                pass
            pass
        pass

    def __contains__(self, key:object) -> bool:
        return key in self.x

    def __delitem__(self, key:K):
        v = self.x[key]
        del self.x[key]
        if self.entered:
            with contextlib.ExitStack() as f:
                f.push(v)
                pass
            pass
        pass

    def __getitem__(self, key:K) -> V:
        return self.x[key]
    
    def __iter__(self):
        return iter(self.x.keys())

    def __len__(self) -> int:
        return len(self.x)

    def __repr__(self) -> str:
        return repr(self.x)

    def __setitem__(self, key:K, value:V) -> None:
        '''replace any current value of {key} with {value}'''
        '''- "enters" the new value and then "exits" any old value'''
        old = self.x.get(key,None)
        if old is not value:
            if self.entered:
                value.__enter__()
                if old is None:
                    self.x[key]=value
                else:
                    with contextlib.ExitStack() as f:
                        f.push(old)
                        self.x[key]=value
                    pass
                pass
            else:
                self.x[key]=value
            pass
        pass
    
    def __sizeof__(self) -> int:
        return self.x.__sizeof__()

    def clear(self):
        while len(self):
            self.popitem()
            pass
        pass

    def get(self, key, default=None):
        return self.x.get(key, default)

    def items(self) -> ItemsView[K,V]:
        '''items in order they were first inserted'''
        return self.x.items()

    def keys(self) -> KeysView[K]:
        '''keys in order they were first inserted'''
        return self.x.keys()

    @overload
    def pop(self, key:K) -> V:
        '''pop and "exit" value of {key} if self has a value for it
           otherwise raise KeyError'''

    @overload
    def pop(self, key:K, default:T) -> Union[V,T]:
        '''pop and "exit" value of {key} if self has a value for it
        otherwise return default'''

    def pop(self, key, default=None):
        if default is None:
            v = self.x.pop(key)
            if self.entered:
                with contextlib.ExitStack() as f:
                    f.push(v)
                pass
            return v
        else:
            try:
                v = self.x.pop(key)
                if self.entered:
                    with contextlib.ExitStack() as f:
                        f.push(v)
                    pass
                return v
            except KeyError:
                return default
            pass
        pass

    def popitem(self) -> Tuple[K, V]:
        '''pop and "exit" most recently added item'''
        k, v = self.x.popitem()
        if self.entered:
            with contextlib.ExitStack() as f:
                f.push(v)
                pass
            pass
        return k, v

    @overload
    def setdefault(self, key:K, default:V) -> V:
        '''set value of {key} to default "entered" if it has no value
           return value of {key}'''

    @overload
    def setdefault(self, key:K, default:None) -> Optional[V]:
        '''same as get(key,None)'''

    def setdefault(self, key, default):
        if default is not None:
            if not key in self:
                self[key] = default
            return self.x.get(key)
        else:
            return self.x.get(key, None)
        pass

    @overload
    def update(self, x:_Dict[K, V]):
        '''update self from {x} assuming its values are not yet entered'''

    @overload
    def update(self, x:Iterable[Tuple[K,V]]):
        '''update self from {x} assuming its values are not yet entered'''

    def update(self, x):
        if getattr(x,'keys',None):
            for k in x.keys():
                self[k] = x[k]
                pass
            pass
        else:
            for k,v in x:
                self[k] = v
                pass
            pass
        pass

class __ClassCm(contextlib.AbstractContextManager):
    '''target object {x} subclass {cls} context management methods
       - so for x: X where X(Y), __ClassCM(Y, x) will call Y.__enter__
         and Y.__exit__ directly avoiding any X.__enter__/__exit__
         overrides'''
    cls:Type
    x:object

    def __init__(self, cls, x):
        self.cls=cls
        self.x=x
        pass
    def __enter__(self):
        self.cls.__enter__(self.x)
        return self
    def __exit__(self, t, e, b):
        return self.cls.__exit__(self.x, t, e, b)
    pass

def is_subclass(n:str, t1, t2:type):
    '''check whether {n}'s type {t1} is a sub-class of {t2}'''
    try:
        Assert(t2).isInstanceOf(type)
        if isinstance(t1,type):
            return issubclass(t1, t2)
        else:
            return False
        pass
    except Exception:
        raise in_function_context(is_subclass,vars()) from None
    pass

ResultType=TypeVar('ResultType')
class Task(contextlib.AbstractAsyncContextManager,Generic[ResultType]):
    '''asyncio Task context manager that guarentees Task awaited before exit
       - note that normally, application will have waited for the task
         within the context; cancel-on-exit ensures clean up on exception
         e.g. when using gather() with its default return_exceptions=False'''
    coroutine: Coroutine[Any,Any,ResultType]
    task: None | asyncio.Task[ResultType]
    def __init__(self, coroutine: Coroutine[Any,Any,ResultType]):
        self.coroutine=coroutine
        self.task=None
    async def __aenter__(self) -> asyncio.Task[ResultType]:
        '''create and return task'''
        assert self.task is None
        self.task=asyncio.create_task(self.coroutine)
        return self.task
    async def __aexit__(self, t, e, b):
        '''cancel and await task'''
        assert self.task is not None
        task=self.task
        self.task=None
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
